from __future__ import annotations

"""
Qdrant Mixins
"""

import abc
import datetime
import typing as t
from lzl import load
from lzl.types import eproperty
from lzo.utils import Timer, parse_datetime
from lzl.logging import Logger, logger as _logger, null_logger

if t.TYPE_CHECKING:
    from pydantic import BaseModel
    from lzl.api.qdrant import QdrantClient, QdrantClientSettings
    from fastembed.rerank.cross_encoder import TextCrossEncoder
    from qdrant_client.conversions import common_types as ct
    from qdrant_client.http import models as qm
    from qdrant_client.fastembed_common import QueryResponse

    class QdrantBaseModel(BaseModel):
        """
        Base Qdrant Model
        """

        @property
        def data_id(self) -> t.Optional[str | int]:
            """
            Returns the data id
            """
            ...


        @classmethod
        def get_qdrant_index_config(cls) -> t.Dict[str, t.Dict[str, t.Any]]:
            """
            Returns the Qdrant index config
            """
            return {}

        def to_qdrant_payload(self, mode: str = 'json', exclude_none: bool = True, context: t.Optional[t.Dict[str, t.Any]] = None, **kwargs) -> t.Tuple[str, str, t.Dict[str, t.Any]]:
            """
            Returns a Qdrant payload
            """
            ...
        
    
QdrantModelT = t.TypeVar('QdrantModelT', bound = 'QdrantBaseModel')

def get_model_config_fields(
    model: t.Type['BaseModel'],
    config_filters: t.Dict[str, t.Union[bool, str, t.Any]] = {
        'qdrant_index': True,
    }, 
    results: t.Literal['included', 'excluded', 'exclusive'] = 'included',
    match_all: bool = True,
) -> t.Dict[str, t.Dict[str, t.Any]]:  # sourcery skip: default-mutable-arg
    """
    Returns the fields for the model
    """
    excluded, included = [], []
    for field_name, field in model.model_fields.items():
        if not field.json_schema_extra: 
            if results == 'exclusive': excluded.append(field_name)
            continue
        for key, value in config_filters.items():
            if key not in field.json_schema_extra: continue
            if field.json_schema_extra[key] == value: 
                included.append(field_name)
            else:
                excluded.append(field_name)
    if match_all: included = [i for i in included if i not in excluded]
    fields = set(excluded if results in {'excluded', 'exclusive'} else included)
    # Now we actually build the config
    config_filter_keys = list(config_filters.keys())
    config: t.Dict[str, t.Dict[str, t.Any]] = {}
    for field_name in fields:
        field_schema = model.model_fields[field_name]
        config[field_name] = {
             k: v for k,v in field_schema.json_schema_extra.items() if k not in 
             config_filter_keys
        }
    return config

_SupportedIndexTypes: t.Set[str] = {'keyword', 'integer', 'float', 'bool', 'geo', 'datetime', 'text', 'uuid'}
CollectionState = t.Literal['create', 'recreate', 'delete']
comp_op = {
    '>': 'gt',
    '<': 'lt',
    '>=': 'gte',
    '<=': 'lte',
}

class QdrantSearchMixin(abc.ABC, t.Generic[QdrantModelT]):
    """
    Qdrant Search Mixin
    """

    collection_name: t.Optional[str] = None
    schema: t.Optional[t.Type[QdrantModelT]] = None

    on_disk_default: t.Optional[bool] = None
    on_disk_by_type: t.Optional[t.Dict[str, bool]] = None
    on_disk_by_field: t.Optional[t.Dict[str, bool]] = None

    dense_model: t.Optional[str] = None
    sparse_model: t.Optional[str] = None
    rerank_model: t.Optional[str] = None
    debug_enabled: t.Optional[bool] = None
    retries: t.Optional[int] = None

    has_dataset: t.Optional[bool] = None

    ds_batch_size: t.Optional[int] = 25
    ds_log_every: t.Optional[int] = None


    def __init__(
        self,
        collection_name: t.Optional[str] = None,
        schema: t.Optional[t.Type[QdrantModelT]] = None,
        dense_model: t.Optional[str] = None,
        sparse_model: t.Optional[str] = None,
        rerank_model: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        
        on_disk_default: t.Optional[bool] = None,
        on_disk_by_type: t.Optional[t.Dict[str, bool]] = None,
        on_disk_by_field: t.Optional[t.Dict[str, bool]] = None,

        has_dataset: t.Optional[bool] = None,

        **kwargs: t.Any,
    ):
        self._extra: t.Dict[str, t.Any] = {}
        if collection_name is not None: self.collection_name = collection_name
        if schema is not None: self.schema = schema
        if dense_model is not None: self.dense_model = dense_model
        if sparse_model is not None: self.sparse_model = sparse_model
        if rerank_model is not None: self.rerank_model = rerank_model
        if debug_enabled is not None: self.debug_enabled = debug_enabled
        if on_disk_default is not None: self.on_disk_default = on_disk_default
        self.on_disk_by_type = on_disk_by_type if on_disk_by_type is not None else {}
        self.on_disk_by_field = on_disk_by_field if on_disk_by_field is not None else {}
        
        if has_dataset is not None: self.has_dataset = has_dataset
        self._initialized: bool = False
        self._ingested: bool = False
        self._kwargs = kwargs

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return _logger

    @eproperty
    def autologger(self) -> 'Logger':
        """
        Automatic Logger that is enabled in devel mode
        """
        return self.logger if self.debug_enabled else null_logger

    @eproperty
    def client(self) -> 'QdrantClient':
        """
        Returns the Qdrant Client
        """
        from lzl.api.qdrant import QdrantClient
        return QdrantClient(
            set_model = self.dense_model,
            set_sparse_model = self.sparse_model,
            retries = self.retries,
            **self._kwargs,
        )
    

    @eproperty
    def settings(self) -> 'QdrantClientSettings':
        """
        Returns the Qdrant Client Settings
        """
        return self.client.settings
    
    @eproperty
    def has_reranker(self) -> bool:
        """
        Returns whether the Opportunity has a Reranker
        """
        return self.rerank_model is not None

    @eproperty
    def reranker(self) -> t.Optional['TextCrossEncoder']:
        """
        Returns the Reranker if provided
        """
        if self.rerank_model is None: return None
        from fastembed.rerank.cross_encoder import TextCrossEncoder
        return TextCrossEncoder(
            model_name = self.rerank_model,
            lazy_load = True,
        )


    @staticmethod
    def get_model_config_from_fields(
        model: t.Type['BaseModel'],
        config_filters: t.Dict[str, t.Union[bool, str, t.Any]] = {
            'qdrant_index': True,
        }, 
        results: t.Literal['included', 'excluded', 'exclusive'] = 'included',
        match_all: bool = True,
    ) -> t.Dict[str, t.Dict[str, t.Any]]:  # sourcery skip: default-mutable-arg
        """
        Returns the model config for the BaseModel
        """
        return get_model_config_fields(
            model = model,
            config_filters = config_filters,
            results = results,
            match_all = match_all,
        )

    @eproperty
    def qm(self):
        """
        Returns the Qm
        """
        from qdrant_client.http import models
        return models
    
    @eproperty
    def schema_model_config(self) -> t.Dict[str, t.Dict[str, t.Any]]:
        """
        Returns the schema model config
        """
        return self.get_model_config_from_fields(self.schema)

    """
    Downstream Methods
    """

    def get_vectors_config(self, **kwargs):
        """
        Returns the collection vectors config
        """
        return self.client.get_fastembed_vector_params(**kwargs)

    def get_sparse_vectors_config(self, **kwargs):
        """
        Returns the collection sparse vectors config
        """
        return self.client.get_fastembed_sparse_vector_params(**kwargs)
    
    def get_collection_config(self, **kwargs) -> t.Dict[str, t.Any]:
        """
        Returns the collection config
        """
        return {}

    def _parse_kwargs(
        self,
        **kwargs,
    ) -> t.Dict[str, t.Any]:
        """
        Parses the kwargs
        """
        return {k: v for k,v in kwargs.items() if v is not None}

    def get_on_disk_default(
        self,
        index_type: str,
        field_name: t.Optional[str] = None,
    ) -> bool:
        """
        Returns the default on disk value
        """
        if field_name and field_name in self.on_disk_by_field: return self.on_disk_by_field[field_name]
        if index_type in self.on_disk_by_type: return self.on_disk_by_type[index_type]
        return self.on_disk_default
    
    def _on_collection_state(
        self,
        state: CollectionState,
        **kwargs,
    ):
        """
        Hook for Collection State Change
        """
        pass

    async def _aon_collection_state(
        self,
        state: CollectionState,
        **kwargs,
    ):
        """
        Hook for Collection State Change
        """
        pass

    def on_collection_state(
        self,
        state: CollectionState,
        is_async: t.Optional[bool] = True,
        **kwargs,
    ):
        """
        Hook for Collection State Change
        """
        func = self._aon_collection_state if is_async else self._on_collection_state
        return func(
            state = state,
            **kwargs,
        )

    """
    Initialization Methods
    """        

    def get_index_config(
        self,
        index_type: str,
        index_config: t.Dict[str, t.Any],
        field_name: t.Optional[str] = None,
    ) -> 'ct.PayloadSchemaType':
        """
        Returns the index config
        """
        if index_type == 'keyword':
            return self.qm.KeywordIndexParams(
                type = "keyword", 
                **self._parse_kwargs(
                    is_tenant = index_config.get('is_tenant'),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                )
            )

        if index_type == 'integer':
            return self.qm.IntegerIndexParams(
                type = "integer",
                **self._parse_kwargs(
                    lookup = index_config.get('lookup', True),
                    range = index_config.get('range', True),
                    is_principal = index_config.get('is_principal'),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )

        if index_type == 'float':
            return self.qm.FloatIndexParams(
                type = "float",
                **self._parse_kwargs(
                    is_principal = index_config.get('is_principal'),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name))
                ),
            )
        
        if index_type == 'bool':
            return self.qm.BoolIndexParams(
                type = "bool",
                **self._parse_kwargs(
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )

        if index_type == 'geo':
            return self.qm.GeoIndexParams(
                type = "geo",
                **self._parse_kwargs(
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )
        

        if index_type == 'datetime':
            return self.qm.DatetimeIndexParams(
                type = "datetime", 
                **self._parse_kwargs(
                    is_principal = index_config.get('is_principal'),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )
        

        if index_type == 'text':
            return self.qm.TextIndexParams(
                type = "text",
                **self._parse_kwargs(
                    tokenizer = index_config.get('tokenizer', self.qm.TokenizerType.WORD),
                    min_token_len = index_config.get('min_token_len', 2),
                    max_token_len = index_config.get('max_token_len', 40),
                    lowercase = index_config.get('lowercase', True),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )
        
        if index_type == 'uuid':
            return self.qm.UuidIndexParams(
                type = "uuid",
                **self._parse_kwargs(
                    is_tenant = index_config.get('is_tenant'),
                    on_disk = index_config.get('on_disk', self.get_on_disk_default(index_type = index_type, field_name = field_name)),
                ),
            )
        
        raise ValueError(f'Invalid index type: {index_type}')

    def _build_index_from_config(
        self,
        qdrant_index_config: t.Dict[str, t.Dict[str, t.Any]],
        **kwargs,
    ):
        """
        Builds the index from the config
        """
        for field_name, index_config in qdrant_index_config.items():
            index_type = index_config.get('index_type', 'text')
            if index_config.get('nested_field_name'): field_name = index_config['nested_field_name']
            if index_type not in _SupportedIndexTypes:
                self.logger.warning(f'Unsupported index type: {index_type}')
                continue
            if 'index_config' in index_config: index_config = index_config['index_config']
            self.client.create_payload_index(
                collection_name = self.collection_name,
                field_name = field_name,
                field_schema = self.get_index_config(
                    index_type = index_type,
                    index_config = index_config,
                    field_name = field_name,
                )
            )
        
    def _init_collection_index(
        self,
        **kwargs,
    ):
        """
        Initializes the collection index

        - This can be overwritten by a downstream method
        """
        if not self.schema: return
        # model_config = self.get_model_config_from_fields(self.schema)
        self.autologger.info(f'Initializing Collection Index: `|g|{self.collection_name}|e|`', colored = True)
        self._build_index_from_config(self.schema_model_config, **kwargs)
        if hasattr(self.schema, 'get_qdrant_index_config'):
            schema_index_config: t.Dict[str, t.Dict[str, t.Any]] = self.schema.get_qdrant_index_config()
            self._build_index_from_config(schema_index_config, **kwargs)
        

    async def _abuild_index_from_config(
        self,
        qdrant_index_config: t.Dict[str, t.Dict[str, t.Any]],
        **kwargs,
    ):
        """
        Builds the index from the config
        """
        for field_name, index_config in qdrant_index_config.items():
            index_type = index_config.get('index_type', 'text')
            if index_config.get('nested_field_name'): field_name = index_config['nested_field_name']
            if index_type not in _SupportedIndexTypes:
                self.logger.warning(f'Unsupported index type: {index_type}')
                continue
            self.autologger.info(f'Setting up index for `|g|{field_name}|e|` with type `|y|{index_type}|e|`', colored = True)
            if 'index_config' in index_config: index_config = index_config['index_config']
            await self.client.acreate_payload_index(
                collection_name = self.collection_name,
                field_name = field_name,
                field_schema = self.get_index_config(
                    index_type = index_type,
                    index_config = index_config,
                    field_name = field_name,
                )
            )


    async def _ainit_collection_index(
        self,
        **kwargs,
    ):
        """
        Initializes the collection index

        - This can be overwritten by a downstream method
        """
        if not self.schema: return
        # model_config = self.get_model_config_from_fields(self.schema)
        self.autologger.info(f'Initializing Collection Index: `|g|{self.collection_name}|e|`', colored = True)
        await self._abuild_index_from_config(self.schema_model_config, **kwargs)
        if hasattr(self.schema, 'get_qdrant_index_config'):
            schema_index_config: t.Dict[str, t.Dict[str, t.Any]] = self.schema.get_qdrant_index_config()
            self.autologger.info(f'Initializing from Qdrant Schema Config: `|g|{self.collection_name}|e|`: {schema_index_config}', colored = True)
            await self._abuild_index_from_config(schema_index_config, **kwargs)


    def _init_collection(
        self,
        skip_init_index: t.Optional[bool] = None,
        **kwargs,
    ):
        """
        Initializes the collection
        """
        if self._initialized: return
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if self.client.collection_exists(self.collection_name): 
            collection_info = self.client.get_collection(collection_name = self.collection_name)
            self.client.api._validate_collection_info(collection_info)
            self._initialized = True
            return
        self.autologger.info(f'Initializing Collection: `|g|{self.collection_name}|e|`', colored = True)
        self.client.create_collection(
            collection_name = self.collection_name,
            vectors_config = self.get_vectors_config(),
            sparse_vectors_config = self.get_sparse_vectors_config(),
            **self.get_collection_config(**kwargs),
        )
        self.on_collection_state('create', is_async = False)
        if not skip_init_index: self._init_collection_index(**kwargs)
        self._initialized = True


    async def _ainit_collection(
        self,
        skip_init_index: t.Optional[bool] = None,
        **kwargs,
    ):
        """
        Initializes the collection
        """
        if self._initialized: return
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if await self.client.acollection_exists(self.collection_name): 
            collection_info = await self.client.aget_collection(collection_name = self.collection_name)
            self.client.api._validate_collection_info(collection_info)
            self._initialized = True
            return
        self.autologger.info(f'Initializing Collection: `|g|{self.collection_name}|e|`', colored = True)
        await self.client.acreate_collection(
            collection_name = self.collection_name,
            vectors_config = self.get_vectors_config(),
            sparse_vectors_config = self.get_sparse_vectors_config(),
            **self.get_collection_config(**kwargs),
        )
        await self.on_collection_state('create', is_async = True)
        if not skip_init_index: await self._ainit_collection_index(**kwargs)
        self._initialized = True

    def init_collection(
        self,
        skip_init_index: t.Optional[bool] = None,
        is_async: t.Optional[bool] = True,
        **kwargs,
    ):
        """
        Initializes the collection
        """
        func = self._ainit_collection if is_async else self._init_collection
        return func(skip_init_index = skip_init_index, **kwargs)

    def _delete_collection(self, **kwargs):
        """
        Deletes the existing collection
        """
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if not self.client.collection_exists(self.collection_name): 
            self.autologger.warning(f'Collection: `|g|{self.collection_name}|e|` does not exist', colored = True)
            return
        self.autologger.info(f'Deleting Collection: `|g|{self.collection_name}|e|`', colored = True)
        self.client.delete_collection(
            collection_name = self.collection_name,
        )
        self.on_collection_state('deletion', is_async = False)
        self._initialized = False
        self._ingested = False

    async def _adelete_collection(self, **kwargs):
        """
        Deletes the existing collection
        """
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if not await self.client.acollection_exists(self.collection_name): 
            self.autologger.warning(f'Collection: `|g|{self.collection_name}|e|` does not exist', colored = True)
            return
        self.autologger.info(f'Deleting Collection: `|g|{self.collection_name}|e|`', colored = True)
        await self.client.adelete_collection(
            collection_name = self.collection_name,
        )
        await self.on_collection_state('deletion', is_async = True)
        self._initialized = False
        self._ingested = False

    def delete_collection(self, is_async: t.Optional[bool] = True, **kwargs):
        """
        Deletes the existing collection
        """
        func = self._adelete_collection if is_async else self._delete_collection
        return func(**kwargs)

    def _recreate_collection(self, skip_init_index: t.Optional[bool] = None, **kwargs):
        """
        Recreates the existing collection with Upsert Operation
        """
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if self.client.collection_exists(self.collection_name):  
            self.delete_collection(is_async = False, **kwargs)
        self.init_collection(skip_init_index = skip_init_index, is_async = False, **kwargs)
        self.on_collection_state('recreate', is_async = False)

    async def _arecreate_collection(self, skip_init_index: t.Optional[bool] = None, **kwargs):
        """
        Recreates the existing collection with Upsert Operation
        """
        if self.collection_name is None: raise ValueError('Collection Name is None')
        if await self.client.acollection_exists(self.collection_name):  
            await self.delete_collection(is_async = True, **kwargs)
        await self.init_collection(skip_init_index = skip_init_index, is_async = True, **kwargs)
        await self.on_collection_state('recreate', is_async = True)

    def recreate_collection(self, skip_init_index: t.Optional[bool] = None, is_async: t.Optional[bool] = True, **kwargs):
        """
        Recreates the existing collection with an Upsert Operation
        """
        func = self._arecreate_collection if is_async else self._recreate_collection
        return func(skip_init_index = skip_init_index, **kwargs)

    """
    Ingestion Methods
    """

    def dataset_iterator(self, **kwargs) -> t.Generator[t.Tuple[str, str, t.Dict[str, t.Any]], None, None]:
        """
        Returns an iterator for the dataset

        - This should be overwritten by a downstream method
        """
        yield
        raise NotImplementedError
    

    async def dataset_aiterator(self, **kwargs) -> t.AsyncGenerator[t.Tuple[str, str, t.Dict[str, t.Any]], None]:
        """
        Returns an async iterator for the dataset

        - This should be overwritten by a downstream method
        """
        yield
        raise NotImplementedError


    def _build_dataset(
        self,
        batch_size: t.Optional[int] = None,
        max_items: t.Optional[int] = None,
        parallel: t.Optional[int] = None,
        **kwargs,
    ):
        """
        Builds the dataset

        - This should be overwritten by a downstream method
        """
        raise NotImplementedError('This method should be overwritten by a downstream method')

    async def _abuild_dataset(
        self,
        batch_size: t.Optional[int] = None,
        max_items: t.Optional[int] = None,
        parallel: t.Optional[int] = None,
        **kwargs,
    ):
        """
        Builds the dataset

        - This should be overwritten by a downstream method
        """
        raise NotImplementedError('This method should be overwritten by a downstream method')

    def _init_dataset(
        self,
        batch_size: t.Optional[int] = None,
        max_items: t.Optional[int] = None,
        log_every: t.Optional[int] = None,
        parallel: t.Optional[int] = None,
        **kwargs,
    ):
        """
        Initializes the dataset
        """
        if not self.has_dataset: return
        if self._ingested: return
        self.autologger.info(f'Initializing Dataset: `|g|{self.collection_name}|e|`', colored = True)
        batch_size = batch_size or self.ds_batch_size
        log_every = log_every if log_every is not None else self.ds_log_every
        result = self._build_dataset(batch_size = batch_size, max_items = max_items, log_every = log_every, parallel = parallel, **kwargs)
        self._ingested = True
        return result
    
    async def _ainit_dataset(
        self,
        batch_size: t.Optional[int] = None,
        max_items: t.Optional[int] = None,
        log_every: t.Optional[int] = None,
        parallel: t.Optional[int] = None,
        **kwargs,
    ):
        """
        Initializes the dataset
        """
        if not self.has_dataset: return
        if self._ingested: return
        self.autologger.info(f'Initializing Dataset: `|g|{self.collection_name}|e|`', colored = True)
        batch_size = batch_size or self.ds_batch_size
        log_every = log_every if log_every is not None else self.ds_log_every
        result = await self._abuild_dataset(batch_size = batch_size, max_items = max_items, log_every = log_every, parallel = parallel, **kwargs)
        self._ingested = True
        return result
    
    def init_dataset(
        self,
        batch_size: t.Optional[int] = None,
        max_items: t.Optional[int] = None,
        log_every: t.Optional[int] = None,
        parallel: t.Optional[int] = None,
        is_async: t.Optional[bool] = True,
        **kwargs,
    ):
        """
        Ingests the dataset
        """
        func = self._ainit_dataset if is_async else self._init_dataset
        return func(batch_size = batch_size, max_items = max_items, log_every = log_every, parallel = parallel, **kwargs)


    """
    Search Methods
    """

    def _filter_datetime_parse(
        self,
        key: str,
        value: t.Union[str, bool, int, float, datetime.datetime, t.Tuple[datetime.datetime, datetime.datetime]],
        op: t.Optional[str] = 'gt',
    ) -> 'qm.FieldCondition':
        """
        Parses the datetime Filter
        
        Possible Values:
        - str
            - '> 2023-08-16'
            - '< 2023-08-16'
            - '>= 2023-08-16'

        - datetime.datetime: we assume that it is always greater than for now.
            - 2023-08-16T00:00:00
            - 2023-08-16
        
        # - int: we assume that it is always greater than for now.
        #     - 
        """
        range_obj = self.qm.DatetimeRange()
        
        if isinstance(value, tuple):
            start, end = value
            start = parse_datetime(start)
            end = parse_datetime(end)
            opposite_op = 'lt' if op == 'gt' else 'gt'
            setattr(range_obj, f'{op}e', start)
            setattr(range_obj, f'{opposite_op}e', end)
            # setattr(range_obj, 'gte', start)
            # setattr(range_obj, 'lte', end)

        elif isinstance(value, str): 
            if value.startswith('>') or value.startswith('<'):
                operator, value = value.split(' ', 1)
                value = parse_datetime(value.strip())
                setattr(range_obj, comp_op[operator], value)
            else:
                value = parse_datetime(value)
                # setattr(range_obj, 'gt', value)
                setattr(range_obj, op, value)
        elif isinstance(value, datetime.datetime):
            value = parse_datetime(value)
            setattr(range_obj, op, value)
            # setattr(range_obj, 'gt', value)

        
        elif isinstance(value, (int, float)):
            now = datetime.datetime.now(tz = datetime.timezone.utc) - datetime.timedelta(days = value)
            # setattr(range_obj, 'gt', now)
            setattr(range_obj, op, now)

        elif isinstance(value, bool):
            # Assume it's >/< now
            now = datetime.datetime.now(tz = datetime.timezone.utc)
            if value:
                setattr(range_obj, op, now)
            else:
                opposite_op = 'lt' if op == 'gt' else 'gt'
                setattr(range_obj, opposite_op, now)

        return self.qm.FieldCondition(key = key, range = range_obj)
            
    def _filter_numeric_parse(
        self,
        key: str,
        value: t.Union[str, int, float, t.Tuple[float, float]],
        op: t.Optional[str] = 'gt',
    ) -> 'qm.FieldCondition':
        """
        Parses the numeric Filter
        
        Possible Values:
        - str
            - '50 - 100'
            - '< 100'
            - '>= 100'
        
        - int: we assume that it is always greater than for now.
            - 50

        - float: we assume that it is always greater than for now.
            - 50.0
        """
        range_obj = self.qm.Range()
        opposite_op = 'lt' if op == 'gt' else 'gt'
        if isinstance(value, tuple):
            start, end = value
            setattr(range_obj, f'{op}e', start)
            setattr(range_obj, f'{opposite_op}e', end)
            # setattr(range_obj, 'gte', start)
            # setattr(range_obj, 'lte', end)

        elif isinstance(value, str): 
            if value.startswith('>') or value.startswith('<'):
                operator, value = value.split(' ', 1)
                value = float(value.strip())
                setattr(range_obj, comp_op[operator], value)
            elif '-' in value:
                start, end = value.split('-', 1)
                start = float(start.strip())
                end = float(end.strip())
                # setattr(range_obj, 'gte', start)
                # setattr(range_obj, 'lte', end)
                setattr(range_obj, f'{op}e', start)
                setattr(range_obj, f'{opposite_op}e', end)
            else:
                value = float(value)
                # setattr(range_obj, 'gt', value)
                setattr(range_obj, op, value)
        else:
            value = float(value)
            # setattr(range_obj, 'gt', value)
            setattr(range_obj, op, value)
        return self.qm.FieldCondition(key = key, range = range_obj)

            

    def build_filters_for_search(
        self,
        query: t.Optional[str] = None,
        **kwargs,
    ) -> t.Tuple[str, t.Optional['qm.Filter']]:
        """
        Builds the filters for the search
        """
        raise NotImplementedError
        # query = query or ""
        # filters = []
        # min_filters = []

    def rerank_results(
        self,
        query: str,
        results: t.List['QueryResponse'],
    ) -> t.List['QueryResponse']:
        """
        Reranks the results
        """
        # self.logger.info(f'Reranking Results for Query: {query}', prefix = self.rerank_model)
        description_hits = [hit.document for hit in results]
        new_scores = list(self.reranker.rerank(query, description_hits))
        ranking = list(enumerate(new_scores)) #saving document indices
        ranking.sort(key=lambda x: x[1], reverse=True) #sorting them in order of relevance defined by reranker

        # return the rerank the results
        return [results[rank[0]] for rank in ranking]
    
    def _schema_validate(
        self,
        hit: t.Dict[str, t.Any] | 'ct.Record' | 'ct.QueryResponse',
        context: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> QdrantModelT | t.Tuple[QdrantModelT, float]:
        """
        Validates the schema
        """
        result = self.schema.model_validate(hit, context = context)
        if context:
            if context.get('include_scores') and hasattr(hit, 'score'):
                return result, hit.score
            if context.get('include_hits'):
                return result, hit
        return result
    
    def _batch_schema_validate(
        self,
        hits: t.List[t.Dict[str, t.Any] | 'ct.Record' | 'ct.QueryResponse'],
        context: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.List[QdrantModelT] | t.List[t.Tuple[QdrantModelT, float]]:
        """
        Validates the schema
        """
        return [self._schema_validate(hit, context = context) for hit in hits]

    async def _aschema_validate(
        self,
        hit: t.Dict[str, t.Any] | 'ct.Record' | 'ct.QueryResponse',
        context: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> QdrantModelT | t.Tuple[QdrantModelT, float]:
        """
        Validates the schema
        """
        return self._schema_validate(hit, context = context)
    
    async def _abatch_schema_validate(
        self,
        hits: t.List[t.Dict[str, t.Any] | 'ct.Record' | 'ct.QueryResponse'],
        context: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.List[QdrantModelT] | t.List[t.Tuple[QdrantModelT, float]]:
        """
        Validates the schema
        """
        return [await self._aschema_validate(hit, context = context) for hit in hits]

    def _fast_query(
        self,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None, 
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT] | t.List[t.Tuple[QdrantModelT, float]]:
        """
        Does a Fast Search Query using Scroll
        """
        _, filters = self.build_filters_for_search(**filters)
        limit = limit or 10
        cursor = self.client.scroll(
            collection_name = self.collection_name,
            scroll_filter = filters,
            limit = limit,
            with_payload = True,
            timeout = 60 * 60,
        )
        results, _ = cursor
        if results:
            return self._batch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})
            # return [self._schema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
        return []
    
    async def _afast_query(
        self,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None, 
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT] | t.List[t.Tuple[QdrantModelT, float]]:
        """
        Does a Fast Search Query using Scroll
        """
        _, filters = self.build_filters_for_search(**filters)
        limit = limit or 10
        cursor = await self.client.ascroll(
            collection_name = self.collection_name,
            scroll_filter = filters,
            limit = limit,
            with_payload = True,
            timeout = 60 * 60,
        )
        results, _ = cursor
        if results:
            return await self._abatch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})
            # return [await self._aschema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
        return []

    @t.overload
    def fast_query(
        self,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ..., 
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[QdrantModelT]]:
        """
        Does a Fast Search Query using Scroll
        """
        ...
    
    @t.overload
    def fast_query(
        self,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[t.Tuple[QdrantModelT, float]]]:
        """
        Does a Fast Search Query using Scroll and includes scores
        """
        ...


    @t.overload
    def fast_query(
        self,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Does a Fast Search Query using Scroll
        """
        ...
    
    @t.overload
    def fast_query(
        self,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[t.Tuple[QdrantModelT, float]]:
        """
        Does a Fast Search Query using Scroll and includes scores
        """
        ...

    def fast_query(
        self,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        is_async: t.Optional[bool] = True,
        **filters: t.Any,
    ) -> t.List[QdrantModelT] | t.Awaitable[t.List[QdrantModelT]]:
        """
        Does a Fast Search Query using Scroll
        """
        func = self._afast_query if is_async else self._fast_query
        return func(limit = limit, include_scores = include_scores, include_hits = include_hits, context = context, **filters)

    def _query(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries the collection
        """
        query, filters = self.build_filters_for_search(query = query, **filters)
        limit = limit or 10
        results = self.client.query(
            collection_name = self.collection_name,
            query_text = query,
            query_filter = filters,
            limit = limit,
        )
        if results:
            if self.has_reranker:
                results = self.rerank_results(query = query, results = results)
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self._schema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            return self._batch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})
        return []

    async def _aquery(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries the collection
        """
        query, filters = self.build_filters_for_search(query = query, **filters)
        limit = limit or 10
        results = await self.client.aquery(
            collection_name = self.collection_name,
            query_text = query,
            query_filter = filters,
            limit = limit,
        )
        if results:
            if self.has_reranker:
                results = self.rerank_results(query = query, results = results)
            # return [await self._aschema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            return await self._abatch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})

        return []
    

    @t.overload
    def query(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[QdrantModelT]]:
        """
        Queries the collection
        """
        ...
    
    @t.overload
    def query(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[t.Tuple[QdrantModelT, float]]]:
        """
        Queries the collection with scores
        """
        ...

    @t.overload
    def query(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries the collection
        """
        ...
    
    @t.overload
    def query(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[t.Tuple[QdrantModelT, float]]:
        """
        Queries the collection with scores
        """
        ...
    
    def query(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        is_async: t.Optional[bool] = True,
        **filters: t.Any,
    ) -> t.List[QdrantModelT] | t.Awaitable[t.List[QdrantModelT]]:
        """
        Queries the collection
        """
        func = self._aquery if is_async else self._query
        return func(query = query, limit = limit, include_scores = include_scores, include_hits = include_hits, context = context, **filters)

    def _search(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries/Searches the collection
        """
        query, filters = self.build_filters_for_search(query = query, **filters)
        limit = limit or 10
        if query:
            results = self.client.query(
                collection_name = self.collection_name,
                query_text = query,
                query_filter = filters,
                limit = limit,
            )
            if results:
                if self.has_reranker:
                    results = self.rerank_results(query = query, results = results)
        else:
            cursor = self.client.scroll(
                collection_name = self.collection_name,
                scroll_filter = filters,
                limit = limit,
                with_payload = True,
                timeout = 60 * 60,
            )
            results, _ = cursor
        if results: 
            # return [self._schema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            return self._batch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})
        return []

    async def _asearch(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries/Searches the collection
        """
        query, filters = self.build_filters_for_search(query = query, **filters)
        limit = limit or 10
        if query:
            results = await self.client.aquery(
                collection_name = self.collection_name,
                query_text = query,
                query_filter = filters,
                limit = limit,
            )
            if results:
                if self.has_reranker:
                    results = self.rerank_results(query = query, results = results)
        else:
            cursor = await self.client.ascroll(
                collection_name = self.collection_name,
                scroll_filter = filters,
                limit = limit,
                with_payload = True,
                timeout = 60 * 60,
            )
            results, _ = cursor
        if results: 
            # return [await self._aschema_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            # return [self.schema.model_validate(hit, context = {'source': 'qdrant'}) for hit in results]
            return await self._abatch_schema_validate(results, context = {'source': 'qdrant', 'include_scores': include_scores, 'include_hits': include_hits, **(context or {})})
        return []

    
    @t.overload
    def search(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[QdrantModelT]]:
        """
        Queries/Searches the collection
        """
        ...
    
    @t.overload
    def search(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[True] = ...,
        **filters: t.Any,
    ) -> t.Awaitable[t.List[t.Tuple[QdrantModelT, float]]]:
        """
        Queries/Searches the collection with scores
        """
        ...

    @t.overload
    def search(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[False] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[QdrantModelT]:
        """
        Queries/Searches the collection
        """
        ...
    
    @t.overload
    def search(
        self,
        query: t.Optional[str] = ...,
        limit: t.Optional[int] = ...,
        include_scores: t.Literal[True] = ...,
        include_hits: t.Optional[bool] = ...,
        context: t.Optional[t.Dict[str, t.Any]] = ...,
        is_async: t.Literal[False] = ...,
        **filters: t.Any,
    ) -> t.List[t.Tuple[QdrantModelT, float]]:
        """
        Queries/Searches the collection with scores
        """
        ...

    def search(
        self,
        query: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        include_scores: t.Optional[bool] = None,
        include_hits: t.Optional[bool] = None,
        context: t.Optional[t.Dict[str, t.Any]] = None,
        is_async: t.Optional[bool] = True,
        **filters: t.Any,
    ) -> t.List[QdrantModelT] | t.Awaitable[t.List[QdrantModelT]]:
        """
        Queries/Searches the collection
        """
        func = self._asearch if is_async else self._search
        return func(query = query, limit = limit, include_scores = include_scores, include_hits = include_hits, context = context, **filters)

    
    def _points_iterator(
        self,
        ids: t.Optional[t.Iterable['qm.ExtendedPointId']],
        metadata: t.Optional[t.Iterable[dict[str, t.Any]]],
        encoded_docs: t.Iterable[tuple[str, list[float]]],
        ids_accumulator: list,
        sparse_vectors: t.Optional[t.Iterable['ct.SparseVector']] = None,
        callback: t.Optional[t.Callable[[t.Any], None]] = None,
    ) -> t.Iterable['qm.PointStruct']:
        """
        The Points Iterator is a generator that yields PointStruct objects.
        """
        for point in self.client.sapi._points_iterator(
            ids = ids,
            metadata = metadata,
            encoded_docs = encoded_docs,
            ids_accumulator = ids_accumulator,
            sparse_vectors = sparse_vectors,
        ):
            yield point
            if callback: callback(point)
        
    
    def add(
        self,
        documents: t.Iterable[str],
        metadata: t.Optional[t.Iterable[dict[str, t.Any]]] = None,
        ids: t.Optional[t.Iterable['qm.ExtendedPointId']] = None,
        batch_size: int = 32,
        parallel: t.Optional[int] = None,
        callback: t.Optional[t.Callable[[t.Any], None]] = None,
        **kwargs: t.Any,
    ) -> list[t.Union[str, int]]:
        """
        Adds text documents into qdrant collection.
        If collection does not exist, it will be created with default parameters.
        Metadata in combination with documents will be added as payload.
        Documents will be embedded using the specified embedding model.

        If you want to use your own vectors, use `upsert` method instead.

        Args:
            collection_name (str):
                Name of the collection to add documents to.
            documents (Iterable[str]):
                List of documents to embed and add to the collection.
            metadata (Iterable[dict[str, Any]], optional):
                List of metadata dicts. Defaults to None.
            ids (Iterable[models.ExtendedPointId], optional):
                List of ids to assign to documents.
                If not specified, UUIDs will be generated. Defaults to None.
            batch_size (int, optional):
                How many documents to embed and upload in single request. Defaults to 32.
            parallel (Optional[int], optional):
                How many parallel workers to use for embedding. Defaults to None.
                If number is specified, data-parallel process will be used.

        Raises:
            ImportError: If fastembed is not installed.

        Returns:
            List of IDs of added documents. If no ids provided, UUIDs will be randomly generated on client side.

        """
        # Wrapper method for the add method
        if not self._initialized: self.init_collection(is_async = False)
        encoded_docs = self.client.sapi._embed_documents(
            documents=documents,
            embedding_model_name=self.client.sapi.embedding_model_name,
            batch_size=batch_size,
            embed_type="passage",
            parallel=parallel,
        )
        encoded_sparse_docs = None
        if self.client.sapi.sparse_embedding_model_name is not None:
            encoded_sparse_docs = self.client.sapi._sparse_embed_documents(
                documents = documents,
                embedding_model_name = self.client.sapi.sparse_embedding_model_name,
                batch_size = batch_size,
                parallel = parallel,
            )
        inserted_ids: list = []
        points = self._points_iterator(
            ids = ids,
            metadata = metadata,
            encoded_docs = encoded_docs,
            ids_accumulator = inserted_ids,
            sparse_vectors = encoded_sparse_docs,
            callback = callback,
        )
        # points = self.client.sapi._points_iterator(
        #     ids = ids,
        #     metadata = metadata,
        #     encoded_docs = encoded_docs,
        #     ids_accumulator = inserted_ids,
        #     sparse_vectors = encoded_sparse_docs,
        # )
        self.client.sapi.upload_points(
            collection_name = self.collection_name,
            points = points,
            wait = True,
            parallel = parallel or 1,
            batch_size = batch_size,
            **kwargs,
        )
        if ids and len(inserted_ids) != len(ids):
            missing_ids = set(ids) - set(inserted_ids)
            self.logger.warning(f'Some IDs were not inserted: {missing_ids} ({len(missing_ids)} missing)')
        return inserted_ids

    def _apoints_iterator(
        self,
        ids: t.Optional[t.Iterable['qm.ExtendedPointId']],
        metadata: t.Optional[t.Iterable[dict[str, t.Any]]],
        encoded_docs: t.Iterable[tuple[str, list[float]]],
        ids_accumulator: list,
        sparse_vectors: t.Optional[t.Iterable['ct.SparseVector']] = None,
        callback: t.Optional[t.Callable[[t.Any], None]] = None,
    ) -> t.Iterable['qm.PointStruct']:
        """
        [Async Client] The Points Iterator is a generator that yields PointStruct objects.
        """
        for point in self.client.api._points_iterator(
            ids = ids,
            metadata = metadata,
            encoded_docs = encoded_docs,
            ids_accumulator = ids_accumulator,
            sparse_vectors = sparse_vectors,
        ):
            yield point
            if callback: callback(point)

    async def aadd(
        self,
        documents: t.Iterable[str],
        metadata: t.Optional[t.Iterable[dict[str, t.Any]]] = None,
        ids: t.Optional[t.Iterable['qm.ExtendedPointId']] = None,
        batch_size: int = 32,
        parallel: t.Optional[int] = None,
        callback: t.Optional[t.Callable[[t.Any], None]] = None,
        **kwargs: t.Any,
    ) -> list[t.Union[str, int]]:
        """
        Adds text documents into qdrant collection.
        If collection does not exist, it will be created with default parameters.
        Metadata in combination with documents will be added as payload.
        Documents will be embedded using the specified embedding model.

        If you want to use your own vectors, use `upsert` method instead.

        Args:
            collection_name (str):
                Name of the collection to add documents to.
            documents (Iterable[str]):
                List of documents to embed and add to the collection.
            metadata (Iterable[dict[str, Any]], optional):
                List of metadata dicts. Defaults to None.
            ids (Iterable[models.ExtendedPointId], optional):
                List of ids to assign to documents.
                If not specified, UUIDs will be generated. Defaults to None.
            batch_size (int, optional):
                How many documents to embed and upload in single request. Defaults to 32.
            parallel (Optional[int], optional):
                How many parallel workers to use for embedding. Defaults to None.
                If number is specified, data-parallel process will be used.

        Raises:
            ImportError: If fastembed is not installed.

        Returns:
            List of IDs of added documents. If no ids provided, UUIDs will be randomly generated on client side.

        """
        if not self._initialized: await self.init_collection(is_async = True)
        encoded_docs = self.client.api._embed_documents(
            documents=documents,
            embedding_model_name=self.client.api.embedding_model_name,
            batch_size=batch_size,
            embed_type="passage",
            parallel=parallel,
        )
        encoded_sparse_docs = None
        if self.client.api.sparse_embedding_model_name is not None:
            encoded_sparse_docs = self.client.api._sparse_embed_documents(
                documents = documents,
                embedding_model_name = self.client.api.sparse_embedding_model_name,
                batch_size = batch_size,
                parallel = parallel,
            )
        inserted_ids: list = []
        points = self._apoints_iterator(
            ids = ids,
            metadata = metadata,
            encoded_docs = encoded_docs,
            ids_accumulator = inserted_ids,
            sparse_vectors = encoded_sparse_docs,
            callback = callback,
        )
        # points = self.client.api._points_iterator(
        #     ids = ids,
        #     metadata = metadata,
        #     encoded_docs = encoded_docs,
        #     ids_accumulator = inserted_ids,
        #     sparse_vectors = encoded_sparse_docs,
        # )
        self.client.api.upload_points(
            collection_name = self.collection_name,
            points = points,
            wait = True,
            parallel = parallel or 1,
            batch_size = batch_size,
            **kwargs,
        )
        if ids and len(inserted_ids) != len(ids):
            missing_ids = set(ids) - set(inserted_ids)
            self.logger.warning(f'Some IDs were not inserted: {missing_ids} ({len(missing_ids)} missing)')
        # self.logger.info(f'{inserted_ids}', prefix = f'{len(inserted_ids)}', colored = True)
        return inserted_ids

    """
    Setup Methods
    """
    @classmethod
    def _setup_(
        cls: t.Type['QdrantSearcher'],
        # Init Keywords
        collection_name: t.Optional[str] = None,
        schema: t.Optional[t.Type[QdrantModelT]] = None,
        dense_model: t.Optional[str] = None,
        sparse_model: t.Optional[str] = None,
        rerank_model: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        on_disk_default: t.Optional[bool] = None,
        on_disk_by_type: t.Optional[t.Dict[str, bool]] = None,
        on_disk_by_field: t.Optional[t.Dict[str, bool]] = None,
        has_dataset: t.Optional[bool] = None,
        

        # Initializers
        skip_init_collection: t.Optional[bool] = None,
        skip_init_index: t.Optional[bool] = None,
        skip_init_dataset: t.Optional[bool] = None,

        # Ingestion Kwargs
        dataset_batch_size: t.Optional[int] = None,
        dataset_max_items: t.Optional[int] = None,
        dataset_log_every: t.Optional[int] = None,

        # Downstream Kwargs
        init_collection_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        init_dataset_kwargs: t.Optional[t.Dict[str, t.Any]] = None,

        # Model Config
        set_cuda: t.Optional[bool | str] = None,
        set_model_config: t.Optional[t.Dict[str, t.Any]] = None,
        set_sparse_model_config: t.Optional[t.Dict[str, t.Any]] = None,

        **kwargs,        
    ) -> 'QdrantSearcher':
        """
        Sets up the Qdrant Searchable Text Collection
        """

        new = cls(
            collection_name = collection_name,
            schema = schema,
            dense_model = dense_model,
            sparse_model = sparse_model,
            rerank_model = rerank_model,
            debug_enabled = debug_enabled,
            on_disk_default = on_disk_default,
            on_disk_by_type = on_disk_by_type,
            on_disk_by_field = on_disk_by_field,
            has_dataset = has_dataset,
            **kwargs,
        )
        if set_cuda and isinstance(set_cuda, str) and set_cuda.lower() == 'auto':
            from lzo.utils.system import get_torch_device_name
            set_cuda = get_torch_device_name() == 'cuda'
        
        if set_cuda and set_cuda is True:
            if not set_model_config: set_model_config = {}
            set_model_config['cuda'] = set_cuda
            if not set_sparse_model_config: set_sparse_model_config = {}
            set_sparse_model_config['cuda'] = set_cuda

        if set_model_config: new.client.shared_config.set_model_config.update(set_model_config)
        if set_sparse_model_config: new.client.shared_config.set_sparse_model_config.update(set_sparse_model_config)

        if not skip_init_collection: 
            new.init_collection(
                skip_init_index = skip_init_index,
                is_async = False,
                **(init_collection_kwargs or {}),
            )
        
        if new.has_dataset and not skip_init_dataset:
            new.init_dataset(
                batch_size = dataset_batch_size,
                max_items = dataset_max_items,
                log_every = dataset_log_every,
                is_async = False,
                **(init_dataset_kwargs or {}),
            )
        
        return new
    
    @classmethod
    async def _asetup_(
        cls: t.Type['QdrantSearcher'],
        # Init Keywords
        collection_name: t.Optional[str] = None,
        schema: t.Optional[t.Type[QdrantModelT]] = None,
        dense_model: t.Optional[str] = None,
        sparse_model: t.Optional[str] = None,
        rerank_model: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        on_disk_default: t.Optional[bool] = None,
        on_disk_by_type: t.Optional[t.Dict[str, bool]] = None,
        on_disk_by_field: t.Optional[t.Dict[str, bool]] = None,
        has_dataset: t.Optional[bool] = None,

        # Initializers
        skip_init_collection: t.Optional[bool] = None,
        skip_init_index: t.Optional[bool] = None,
        skip_init_dataset: t.Optional[bool] = None,

        # Ingestion Kwargs
        dataset_batch_size: t.Optional[int] = None,
        dataset_max_items: t.Optional[int] = None,
        dataset_log_every: t.Optional[int] = None,

        # Downstream Kwargs
        init_collection_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        init_dataset_kwargs: t.Optional[t.Dict[str, t.Any]] = None,

        # Model Config
        set_cuda: t.Optional[bool | str] = None,
        set_model_config: t.Optional[t.Dict[str, t.Any]] = None,
        set_sparse_model_config: t.Optional[t.Dict[str, t.Any]] = None,

        **kwargs,        
    ) -> 'QdrantSearcher':
        """
        Sets up the Qdrant Searchable Text Collection
        """

        new = cls(
            collection_name = collection_name,
            schema = schema,
            dense_model = dense_model,
            sparse_model = sparse_model,
            rerank_model = rerank_model,
            debug_enabled = debug_enabled,
            on_disk_default = on_disk_default,
            on_disk_by_type = on_disk_by_type,
            on_disk_by_field = on_disk_by_field,
            has_dataset = has_dataset,
            **kwargs,
        )
        if set_cuda and isinstance(set_cuda, str) and set_cuda.lower() == 'auto':
            from lzo.utils.system import get_torch_device_name
            set_cuda = get_torch_device_name() == 'cuda'
        
        if set_cuda and set_cuda is True:
            if not set_model_config: set_model_config = {}
            set_model_config['cuda'] = set_cuda
            if not set_sparse_model_config: set_sparse_model_config = {}
            set_sparse_model_config['cuda'] = set_cuda

        if set_model_config: new.client.shared_config.set_model_config.update(set_model_config)
        if set_sparse_model_config: new.client.shared_config.set_sparse_model_config.update(set_sparse_model_config)

        if not skip_init_collection: 
            await new.init_collection(
                skip_init_index = skip_init_index,
                is_async = True,
                **(init_collection_kwargs or {}),
            )
        
        if new.has_dataset and not skip_init_dataset:
            await new.init_dataset(
                batch_size = dataset_batch_size,
                max_items = dataset_max_items,
                log_every = dataset_log_every,
                is_async = True,
                **(init_dataset_kwargs or {}),
            )
        return new
    
    @classmethod
    def setup(
        cls: t.Type['QdrantSearcher'],
        # Init Keywords
        collection_name: t.Optional[str] = None,
        schema: t.Optional[t.Type[QdrantModelT]] = None,
        dense_model: t.Optional[str] = None,
        sparse_model: t.Optional[str] = None,
        rerank_model: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        on_disk_default: t.Optional[bool] = None,
        on_disk_by_type: t.Optional[t.Dict[str, bool]] = None,
        on_disk_by_field: t.Optional[t.Dict[str, bool]] = None,
        has_dataset: t.Optional[bool] = None,

        # Initializers
        skip_init_collection: t.Optional[bool] = None,
        skip_init_index: t.Optional[bool] = None,
        skip_init_dataset: t.Optional[bool] = None,

        # Ingestion Kwargs
        dataset_batch_size: t.Optional[int] = None,
        dataset_max_items: t.Optional[int] = None,
        dataset_log_every: t.Optional[int] = None,

        # Downstream Kwargs
        init_collection_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        init_dataset_kwargs: t.Optional[t.Dict[str, t.Any]] = None,

        # Model Config
        set_cuda: t.Optional[bool | str] = None,
        set_model_config: t.Optional[t.Dict[str, t.Any]] = None,
        set_sparse_model_config: t.Optional[t.Dict[str, t.Any]] = None,

        is_async: t.Optional[bool] = True,
        **kwargs,        
    ) -> 'QdrantSearcher' | t.Awaitable['QdrantSearcher']:
        """
        Sets up the Qdrant Searchable Text Collection
        """
        method = cls._asetup_ if is_async else cls._setup_
        return method(
            collection_name = collection_name,
            schema = schema,
            dense_model = dense_model,
            sparse_model = sparse_model,
            rerank_model = rerank_model,
            debug_enabled = debug_enabled,
            on_disk_default = on_disk_default,
            on_disk_by_type = on_disk_by_type,
            on_disk_by_field = on_disk_by_field,
            has_dataset = has_dataset,

            # Initializers
            skip_init_collection = skip_init_collection,
            skip_init_index = skip_init_index,
            skip_init_dataset = skip_init_dataset,

            # Ingestion Kwargs
            dataset_batch_size = dataset_batch_size,
            dataset_max_items = dataset_max_items,
            dataset_log_every = dataset_log_every,

            # Downstream Kwargs
            init_collection_kwargs = init_collection_kwargs,
            init_dataset_kwargs = init_dataset_kwargs,

            # Model Config
            set_cuda = set_cuda,
            set_model_config = set_model_config,
            set_sparse_model_config = set_sparse_model_config,
            **kwargs,
        )

    
QdrantSearcher = t.TypeVar('QdrantSearcher', bound = QdrantSearchMixin)


"""
Qdrant Client with Unified Async / Sync 
"""

from .config import QdrantClientSettings, settings

# We need to use the full client because it waits for models to be added before initializing
from .full_client import QdrantClient


from __future__ import annotations

import time
import secrets
import typing as t
from pydantic.networks import AnyHttpUrl
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    Protocol,
    cast,
)

from . import logging

# import requests
from .cursor import Cursor
from .result import ResultItem
from .utils import parse_host
from .client import ConnectionClient, ConnURL

class SyncWritableIO(Protocol):
    def write(self, data: bytes, /) -> Any:
        ...


class AsyncWritableIO(Protocol):
    async def write(self, data: bytes, /) -> Any:
        ...


T = TypeVar("T")


class Connection:
    """Describes a synchronous description of a pool of nodes which
    make up a rqlite cluster. There is no actual connection in this
    class, however the name is chosed to signify its similarity to
    a Connection in a DBAPI 2.0 style api.
    """

    def __init__(
        self,
        hosts: Union[Union[str, AnyHttpUrl], List[Union[str, AnyHttpUrl]]],
        timeout: int = 5,
        max_redirects: int = 2,
        max_attempts_per_host: int = 2,
        read_consistency: Literal["none", "weak", "strong"] = "weak",
        freshness: str = "0",
        log: Union[logging.LogConfig, bool] = True,
        default_mode: Literal["sync", "async"] = "sync",
        default_return: Literal["result", "cursor"] = "result",
    ):
        """Initializes a new synchronous Connection. This is typically
        called with the alias rqlite.connect

        Args:
            hosts (list[str]): A list of hostnames or ip addresses where
                the rqlite cluster is running. Each address can be a
                host:port pair - if the port is omitted, the default
                port of 4001 is used. If the host is specified in ipv6
                format, then it should be surrounded in square brackets
                if the port is specified, i.e., [::1]:4001.

            timeout (int): The number of seconds to wait for a response
                from a host before giving up.

            max_redirects (int): The maximum number of redirects to follow
                when executing a query.

            max_attempts_per_host (int): The maximum number of attempts to
                make to a given host before giving up.

            read_consistency (str): The read consistency to use for cursors
                by default.

            freshness (str): Default freshness value for none-consistency reads.
                Guarrantees the response is no more stale than the given duration,
                specified as a golang Duration string, e.g., "5m" for 5 minutes

            log (bool or LogConfig): If True, logs will have the default settings.
                If False, logs will be disabled. If a LogConfig, the
                configuration of the logs.

            default_mode (Literal["sync", "async"]): The default mode to use for
        """
        if log is True:
            log_config = logging.LogConfig()
        elif log is False:
            log_config = logging.DISABLED_LOG_CONFIG
        else:
            log_config = log

        # self.hosts: List[Tuple[str, int]] = [parse_host(h) for h in hosts]
        if not isinstance(hosts, list): hosts = [hosts]
        self.hosts: List[ConnURL] = [ConnURL(host = parse_host(h)) for h in hosts]
        """The host addresses to connect to as a list of (hostname, port) pairs."""

        self.timeout: int = timeout
        """The number of seconds to wait for a response from a node."""

        self.max_redirects: int = max_redirects
        """The maximum number of redirects to follow when executing a query."""

        self.max_attempts_per_host: int = max_attempts_per_host
        """The maximum number of attempts to make to a node before giving up. We
        will attempt all nodes in a random order this many times per node before
        giving up.
        """

        self.read_consistency: Literal["none", "weak", "strong"] = read_consistency
        """The default read consistency when initializing cursors."""

        self.freshness: str = freshness
        """Default freshness value for none-consistency reads."""

        self.log_config = log_config
        """The log configuration for this connection and all cursors it creates."""
        self.io = ConnectionClient(
            hosts = self.hosts,
            log_config = self.log_config,
            timeout = self.timeout,
            max_redirects = self.max_redirects,
            max_attempts_per_host = self.max_attempts_per_host,
        )
        self.default_mode = default_mode
        self.default_return = default_return
    
    @property
    def _is_async_default(self) -> bool:
        return self.default_mode == "async"

    def cursor(
        self,
        read_consistency: Optional[Literal["none", "weak", "strong"]] = None,
        freshness: Optional[str] = None,
        default_mode: Literal["sync", "async"] = None,
        default_return: Literal["result", "cursor"] = None,
    ) -> "Cursor":
        """Creates a new cursor for this connection.

        Args:
            read_consistency: The read consistency to use for this cursor. If
                None, the default read consistency for this connection will be
                used.
            freshness: The freshness value to use for this cursor. If None, the
                default freshness value for this connection will be used.

        Returns:
            A new cursor for this connection.
        """
        if read_consistency is None: read_consistency = self.read_consistency
        if freshness is None: freshness = self.freshness
        if default_mode is None: default_mode = self.default_mode
        if default_return is None: default_return = self.default_return
        return Cursor(self, read_consistency, freshness, default_mode, default_return = default_return)

    def backup(self, file: SyncWritableIO, /, raw: bool = False) -> None:
        """Backup the database to a file.

        Args:
            file (file-like): The file to write the backup to.
            raw (bool): If true, the backup will be in raw SQL format. If false, the
                backup will be in the smaller sqlite format.
        """
        request_id = secrets.token_hex(4)
        logging.log(
            self.log_config.backup_start,
            lambda _: f"  [RQLITE BACKUP {{{request_id}}}] raw={raw}",
        )
        request_started_at = time.perf_counter()
        # It is much faster for us to discover the leader and fetch the backup
        # from the leader right now: https://github.com/rqlite/rqlite/issues/1551
        leader = self.io.discover_leader()

        if raw:
            resp = self.io.fetch_response(
                "GET", "/db/backup?fmt=sql", stream=True, initial_host=leader
            )
        else:
            resp = self.io.fetch_response(
                "GET", "/db/backup", stream=True, initial_host=leader
            )

        for chunk in resp.iter_content(chunk_size=4096):
            file.write(chunk)

        resp.close()
        time_taken = time.perf_counter() - request_started_at
        logging.log(
            self.log_config.backup_end,
            lambda _: f"    {{{request_id}}} in {time_taken:.3f}s ->> backup fully written",
        )

    async def abackup(self, file: AsyncWritableIO, /, raw: bool = False) -> None:
        """Backup the database to a file.
        Args:
            file (file-like): The file to write the backup to.
            raw (bool): If true, the backup will be in raw SQL format. If false, the
                backup will be in the smaller sqlite format.
        """
        request_id = secrets.token_hex(4)
        logging.log(
            self.log_config.backup_start,
            lambda _: f"  [RQLITE BACKUP {{{request_id}}}] raw={raw}",
        )
        request_started_at = time.perf_counter()
        # It is much faster for us to discover the leader and fetch the backup
        # from the leader right now: https://github.com/rqlite/rqlite/issues/1551
        leader = await self.io.adiscover_leader()

        if raw:
            resp = await self.io.afetch_response(
                "GET", "/db/backup?fmt=sql", stream=True, initial_host=leader
            )
        else:
            resp = await self.io.afetch_response(
                "GET", "/db/backup", stream=True, initial_host=leader
            )

        async for chunk in resp.iter_content(chunk_size=4096):
            await file.write(chunk)

        await resp.close()
        time_taken = time.perf_counter() - request_started_at
        logging.log(
            self.log_config.backup_end,
            lambda _: f"    {{{request_id}}} in {time_taken:.3f}s ->> backup fully written",
        )

    def execute(
        self,
        operation: str,
        parameters: t.Optional[t.Iterable[t.Any]] = None,
        raise_on_error: bool = True,
        read_consistency: t.Optional[t.Literal["none", "weak", "strong"]] = None,
        freshness: t.Optional[str] = None,
        is_async: t.Optional[bool] = None,
        return_type: t.Optional[Literal["result", "cursor"]] = None,
    ) -> t.Union[ResultItem, t.Awaitable[ResultItem]]:
        """Executes a single query and returns the result. This will also
        update this object so that fetchone() and related functions can be used
        to fetch the results instead.

        Args:
            operation (str): The query to execute.
            parameters (iterable): The parameters to pass to the query.
            raise_on_error (bool): If True, raise an error if the query fails. If
                False, you can check the result item's error property to see if
                the query failed.
            read_consistency (Optional[Literal["none", "weak", "strong"]]):
                The read consistency to use when executing the query. If None,
                use the default read consistency for this cursor.
            freshness (Optional[str]): The freshness to use when executing
                none read consistency queries. If None, use the default freshness
                for this cursor.

        Returns:
            ResultItem: The result of the query.
        """
        if is_async is None: is_async = self._is_async_default
        if return_type is None: return_type = self.default_return
        return self.cursor(
            read_consistency = read_consistency, 
            freshness = freshness,
            default_return = return_type,
        ).execute(
            operation = operation,
            parameters = parameters,
            raise_on_error = raise_on_error,
            is_async = is_async,
        )



from __future__ import annotations

import typing as t

from lzl.api import aioreq
from pydantic.networks import AnyHttpUrl
from pyrqlite.extensions import PARSE_DECLTYPES, PARSE_COLNAMES
from urllib.parse import urlparse
from .utils import parse_host, logger


class Connection(object):
    """
    Describes a synchronous description of a pool of nodes which
    make up a rqlite cluster. There is no actual connection in this
    class, however the name is chosed to signify its similarity to
    a Connection in a DBAPI 2.0 style api.
    """

    def __init__(
        self,
        url: t.Union[str, AnyHttpUrl],
        timeout: t.Optional[float] = None,
        detect_types: t.Optional[bool] = True,
        max_redirects: t.Optional[int] = -1,
        max_retries: t.Optional[int] = 10,
        read_consistency: t.Literal["none", "weak", "strong"] = "strong",
        freshness: t.Optional[str] = "0", 
        default_mode: t.Literal["sync", "async"] = "sync",
    ):
        """
        Initializes a new synchronous Connection. This is typically
        called with the alias aiorqlite.connect

        Args:
            url: The url of the rqlite cluster
        
            timeout (int): The number of seconds to wait for a response
                from a host before giving up.

            max_redirects (int): The maximum number of redirects to follow
                when executing a query.

            read_consistency (str): The read consistency to use for cursors
                by default.

            freshness (str): Default freshness value for none-consistency reads.
                Guarrantees the response is no more stale than the given duration,
                specified as a golang Duration string, e.g., "5m" for 5 minutes

            default_mode (Literal["sync", "async"]): The default mode to use for
                cursors.
        """
        self.messages = []
        self._ephemeral = None
        if url == ':memory:':
            from pyrqlite._ephemeral import EphemeralRqlited
            self._ephemeral = EphemeralRqlited().__enter__()
            host, port = self._ephemeral.http
            url = f'http://{host}:{port}'

        self.url = parse_host(url)
        self.timeout = timeout
        self.max_redirects = max_redirects
        self.max_retries = max_retries
        self.read_consistency = read_consistency
        self.freshness = freshness
        self.default_mode = default_mode

        self.detect_types = detect_types
        self.parse_decltypes = detect_types & PARSE_DECLTYPES
        self.parse_colnames = detect_types & PARSE_COLNAMES
        self.io: aioreq.Client = self._init_connection()

    
    def _init_connection(self) -> aioreq.Client:
        """
        Initializes the Connection
        """
        kwargs = {
            'base_url': f'{self.url.scheme}://{self.url.host}',
            'timeout': None if self.timeout is None else float(self.timeout)
        }
        if self.url.port:
            kwargs['base_url'] += f':{self.url.port}'
        if self.url.username:
            from niquests.auth import HTTPBasicAuth
            kwargs['auth'] = HTTPBasicAuth(self.url.username, self.url.password)
        return aioreq.Client(**kwargs)


    def _retry_request(self, method: str, uri: str, body: t.Optional[t.Any] = None, headers: t.Optional[t.Dict[str, str]] = None) -> 'aioreq.Response':
        """
        [Sync] Retries a request
        """
        tries = self.max_retries
        headers = headers or None
        while tries:
            tries -= 1
            try:
                return self._connection.request(method, uri, data=body, headers=headers)
            except Exception as e:
                if not tries: raise e
                self._connection.close()
                self._connection = self._init_connection()

    async def _aretry_request(self, method: str, uri: str, body: t.Optional[t.Any] = None, headers: t.Optional[t.Dict[str, str]] = None) -> 'aioreq.Response':
        """
        [Async] Retries a request
        """
        tries = self.max_retries
        headers = headers or None
        while tries:
            tries -= 1
            try:
                return await self._aconnection.arequest(method, uri, data=body, headers=headers)
            except Exception as e:
                if not tries: raise e
                await self._aconnection.aclose()
                self._aconnection = self._init_connection()

    
    def _fetch_response(self, method: str, uri: str, body: t.Optional[t.Any] = None, headers: t.Optional[t.Dict[str, str]] = None) -> 'aioreq.Response':
        """
        [Sync] Fetch a response, handling redirection.
        """
        response = self._retry_request(method, uri, body=body, headers=headers)
        redirects = 0
        while response.status_code == 301 and \
                response.headers.get('Location') is not None and \
                (self.max_redirects == -1 or redirects < self.max_redirects):
            
            redirects += 1
            uri = response.headers.get('Location')
            location = urlparse(uri)
            logger.info(f'Status: {response.status_code} Reason: {response.reason} Location: {uri}')
            if self.url.host != location.hostname or self.url.port != location.port:
                self._connection.close()
                new_url = f'{location.scheme}://{self.url.username}:{self.url.password}@{location.hostname}:{location.port}' if self.url.password else f'{location.scheme}://{location.hostname}:{location.port}'
                self.url = parse_host(new_url)
                self._connection = self._init_connection()

            response = self._retry_request(method, uri, body=body, headers=headers)
        return response
    
    async def _afetch_response(self, method: str, uri: str, body: t.Optional[t.Any] = None, headers: t.Optional[t.Dict[str, str]] = None) -> 'aioreq.Response':
        """
        [Async] Fetch a response, handling redirection.
        """
        response = await self._aretry_request(method, uri, body=body, headers=headers)
        redirects = 0
        while response.status_code == 301 and \
                response.headers.get('Location') is not None and \
                (self.max_redirects == -1 or redirects < self.max_redirects):
            
            redirects += 1
            uri = response.headers.get('Location')
            location = urlparse(uri)
            logger.info(f'Status: {response.status_code} Reason: {response.reason} Location: {uri}')
            if self.url.host != location.hostname or self.url.port != location.port:
                self._connection.close()
                new_url = f'{location.scheme}://{self.url.username}:{self.url.password}@{location.hostname}:{location.port}' if self.url.password else f'{location.scheme}://{location.hostname}:{location.port}'
                self.url = parse_host(new_url)
                self._connection = self._init_connection()

            response = await self._aretry_request(method, uri, body=body, headers=headers)
        return response
    

    def close(self):
        """
        Close the connection now (rather than whenever .__del__() is
        called).

        The connection will be unusable from this point forward; an
        Error (or subclass) exception will be raised if any operation
        is attempted with the connection. The same applies to all
        cursor objects trying to use the connection. Note that closing
        a connection without committing the changes first will cause an
        implicit rollback to be performed."""
        self._connection.close()
        if self._ephemeral is not None:
            self._ephemeral.__exit__(None, None, None)
            self._ephemeral = None

    def __del__(self):
        self.close()

    def commit(self):
        """Database modules that do not support transactions should
        implement this method with void functionality."""
        pass

    def rollback(self):
        """This method is optional since not all databases provide
        transaction support. """
        pass

    def cursor(self, factory=None):
        """Return a new Cursor Object using the connection."""
        if factory:
            return factory(self)
        else:
            return Cursor(self)

    def execute(self, *args, **kwargs):
        return self.cursor().execute(*args, **kwargs)

    def ping(self, reconnect=True):
        if self._connection.sock is None:
            if reconnect:
                self._connection = self._init_connection()
            else:
                raise self.Error("Already closed")
        try: 
            self.execute("SELECT 1")
        except Exception:
            if reconnect:
                self._connection = self._init_connection()
                self.ping(False)
            else:
                raise


from __future__ import annotations

"""
HTTP / DNS Resolver Utilities
"""

import time
import socket
import functools
import contextlib
from urllib.parse import urlparse
from lzl import load
from typing import Optional, List, Dict, TYPE_CHECKING

if load.TYPE_CHECKING:
    import pythonping
    import tldextract
    import async_lru
    from lzl.api import aiohttpx
else:
    pythonping = load.LazyLoad("pythonping")
    aiohttpx = load.lazy_load('lzl.api.aiohttpx', install_missing = False)
    tldextract = load.LazyLoad("tldextract")
    async_lru = load.LazyLoad("async_lru")


def normalize_url(url: str) -> str:
    """
    Normalizes the URL
    """
    url = url.lower()
    url = url.replace('http//', '').replace('https//', '').replace('htpps://', '').replace('htp://', '')
    if '@' in url: url = url.split('@')[-1]
    if url.startswith('www://'): url = url.replace('www://', 'www.')
    if not url.startswith('http://') and not url.startswith('https://'): url = f'https://{url}'
    return url.replace('http://', 'https://').rstrip('/')


@functools.lru_cache(500)
def extract_registered_domain(url: str) -> str:
    """
    Returns the domain
    """
    return tldextract.extract(url).registered_domain

def extract_clean_domain(
    url: str
) -> str:
    """
    Extracts and cleans the domain
    """
    domain = extract_registered_domain(url.lower())
    if domain.startswith('www.'): domain = domain.replace('www.', '')
    return domain.lower()

_http_download_headers: Optional[Dict[str, str]] = None

def get_http_download_headers() -> Dict[str, str]:
    """
    Returns the http download headers
    """
    global _http_download_headers
    if _http_download_headers is None:
        try:
            from browserforge.headers import HeaderGenerator
            _http_download_headers = HeaderGenerator().generate(browser='chrome')
        except ImportError:
            _http_download_headers = {
                'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36',
                'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
                'accept-language': 'en-US,en;q=0.9',
                'accept-encoding': 'gzip, deflate, br',
                'cache-control': 'no-cache',
            }
    return _http_download_headers



"""
Validators
"""

def validate_website_with_ping(
    url: str,
    timeout: Optional[int] = 2,
    count: Optional[int] = 4,
) -> bool:
    """
    Validates a website
    """
    url = normalize_url(url)
    ping_results = pythonping.ping(url, count = count, timeout = timeout)
    return ping_results.success(3)


def validate_website_with_socket(
    url: str,
) -> bool:
    """
    Validates a website
    """
    url = extract_clean_domain(url)
    with contextlib.suppress(Exception):
        socket.gethostbyname(url)
        return True
    return False

def validate_website_with_httpx(
    url: str,
    timeout: Optional[int] = 15,
    headers: Optional[Dict[str, str]] = None,
    **kwargs,
) -> bool:
    """
    Validates a website
    """
    url = normalize_url(url)
    try:
        # Try with head first
        headers = headers or get_http_download_headers()
        response = aiohttpx.head(url, timeout = timeout, headers = headers,follow_redirects = True)
        if response.status_code == 405:
            response = aiohttpx.get(url, timeout = timeout, headers = headers, follow_redirects = True)
        response.raise_for_status()
        return True
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return False
    except aiohttpx.HTTPStatusError as e:
        return e.response.status_code < 500
    except Exception as e:
        return False

async def avalidate_website_with_httpx(
    url: str,
    timeout: Optional[int] = 15,
    headers: Optional[Dict[str, str]] = None,
    **kwargs,
) -> bool:
    """
    Validates a website
    """
    url = normalize_url(url)
    try:
        # Try with head first
        headers = headers or get_http_download_headers()
        response = await aiohttpx.async_head(url, timeout = timeout, headers = headers, follow_redirects = True)
        if response.status_code == 405:
            response = await aiohttpx.async_get(url, headers = headers, timeout = timeout, follow_redirects = True)
        response.raise_for_status()
        return True
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return False
    except aiohttpx.HTTPStatusError as e:
        return e.response.status_code < 500
    except Exception as e:
        return False

@functools.lru_cache(maxsize=1200)
def validate_hostname(url: str) -> bool:
    """
    Validates the hostname
    """    
    hostname = urlparse(url).hostname if '://' in url else url
    for _attempts in range(5):
        with contextlib.suppress(Exception):
            socket.gethostbyname(hostname)
            return True
        time.sleep(1.5)
    return False


def validate_website_exists(
    url: str,
    timeout: Optional[int] = 15,
    headers: Optional[Dict[str, str]] = None,
    soft_validate: Optional[bool] = False,
) -> bool:
    """
    Validates the website exists
    """
    url = normalize_url(url)
    hn_valid = validate_website_with_ping(url)
    if soft_validate or not hn_valid: return hn_valid
    return validate_website_with_httpx(url, headers = headers, timeout = timeout)


async def avalidate_website_exists(
    url: str,
    timeout: Optional[int] = 15,
    headers: Optional[Dict[str, str]] = None,
    soft_validate: Optional[bool] = False,
) -> bool:
    """
    Validates the website exists
    """
    url = normalize_url(url)
    hn_valid = validate_website_with_ping(url)
    if soft_validate or not hn_valid: return hn_valid
    return await avalidate_website_with_httpx(url, headers = headers, timeout = timeout)


"""
Resolvers
"""


@functools.lru_cache()
def resolve_domain(url: str, attempts: Optional[int] = None) -> str:
    """
    Returns the root domain of a website after redirection
    """
    if url.startswith('www://'): url = url.replace('www://', '')
    if not url.startswith('http'): url = f'https://{url}'
    try:
        with aiohttpx.Client(follow_redirects = True, verify = False) as client:
            r = client.get(url, timeout = 5)
        r.raise_for_status()
        return str(r.url).rstrip('/')
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return url
    except aiohttpx.HTTPStatusError as e:
        return str(e.response.url).rstrip('/') if e.response.status_code < 500 else url
    except Exception as e:
        attempts = attempts + 1 if attempts else 1
        if attempts > 2:
            return None
        new_url = f'https://{extract_registered_domain(url)}'
        return resolve_domain(new_url, attempts = attempts)
    


@async_lru.alru_cache(maxsize=1200)
async def aresolve_domain(url: str, attempts: Optional[int] = None) -> str:
    """
    Returns the root domain after redirection
    """
    if url.startswith('www://'): url = url.replace('www://', '')
    if not url.startswith('http'): url = f'https://{url}'
    try:
        async with aiohttpx.Client(follow_redirects = True, verify = False) as client:
            r = await client.async_get(url, timeout = 5)
        r.raise_for_status()
        return str(r.url).rstrip('/')
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return url
    except aiohttpx.HTTPStatusError as e:
        return str(e.response.url).rstrip('/') if e.response.status_code < 500 else url
    except Exception as e:
        attempts = attempts + 1 if attempts else 1
        if attempts > 2:
            return None
        new_url = f'https://{extract_registered_domain(url)}'
        return await aresolve_domain(new_url, attempts = attempts)

def determine_invalid_domains(
    urls: List[str],
    timeout: Optional[int] = 4,
) -> List[str]:
    """
    Determines the invalid domains
    """
    if not urls: return []
    return [
        url
        for url in urls
        if not validate_website_with_ping(url, timeout=timeout)
    ]





import json
import contextlib
from lzl.api import aiohttpx
from pydantic import PrivateAttr
from typing import Any, Optional, Union, Dict
from lzo.types import BaseModel, eproperty

# from lazyops.types import BaseModel
# from lazyops.libs.abcs.types.props import eproperty

__all__ = [
    "OpenAIError",
    "ExceptionModel",
    "fatal_exception",
    "APIError",
    "TryAgain",
    "APIConnectionError",
    "Timeout",
    "InvalidRequestError",
    "AuthenticationError",
    "PermissionError",
    "RateLimitError",
    "ServiceUnavailableError",
    "InvalidAPIType",
    "error_handler",
]

class ExceptionModel(BaseModel):
    response: aiohttpx.Response
    data: Optional[Union[Dict, Any]]
    message: Optional[str] = None
    should_retry: Optional[bool] = False
    _extra: Dict[str, Any] = PrivateAttr(default_factory = dict)

    @eproperty
    def headers(self):
        """
        Returns the response headers.
        """
        return self.response.headers
    
    @eproperty
    def stream(self) -> bool:
        """
        Returns True if the response is a streaming response.
        """
        return "text/event-stream" in self.headers.get("content-type", "")

    @eproperty
    def response_data(self):
        return self.data or self.response.json()
    
    @eproperty
    def http_body(self):
        body = self.data if self.stream else self.response.content
        try:
            if hasattr(body, "decode"):
                body = body.decode("utf-8")
            return body
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            raise ValueError(
                f"HTTP code {self.status_code} from API ({body})"
            ) from e
    
    @eproperty
    def response_json(self):
        try:
            return json.loads(self.http_body)
        except json.JSONDecodeError:
            return {}

    @eproperty
    def response_text(self):
        return self.response.text
    
    @eproperty
    def status_code(self):
        return self.response.status_code
    
    @eproperty
    def error_data(self) -> Dict:
        return self.response_json.get("error", {})
    
    @eproperty
    def request_id(self) -> str:
        return self.headers.get("request-id", None)
    
    @eproperty
    def organization(self) -> str:
        return self.headers.get("openai-organization", None)
    
    @eproperty
    def error_message(self) -> str:
        msg: str = self.message or ("(Error occurred while streaming.)" if self.stream else "")
        if self.error_data.get("message"):
            msg += " " + self.error_data.get("message")
        if self.error_data.get("internal_message"):
            msg += "\n\n" + self.error_data["internal_message"]
        return msg.strip() or self.response_text


class OpenAIError(Exception):
    def __init__(
        self,
        response: aiohttpx.Response,
        data: Optional[Union[Dict, Any]],
        message: Optional[str] = None,
        should_retry: Optional[bool] = False,
        **kwargs
    ):
        self.status = response.status_code
        self.response = response
        self.message = message
        self.exc = ExceptionModel(
            response=response,
            message=message,
            data=data,
            should_retry=should_retry,
            **kwargs
        )
        self.post_init(**kwargs)
    
    def post_init(self, **kwargs):
        pass
    
    def __str__(self):
        msg = self.exc.error_message or "<empty message>"
        if self.exc.request_id is not None:
            return f"Request {self.exc.request_id}: {msg}"
        else:
            return msg
        
    @property
    def user_message(self):
        return self.exc.error_message
    
    def __repr__(self):
        return f"[OpenAI] {self.__class__.__name__} \
            (message={self.exc.error_message}, \
                http_status={self.exc.status_code}, \
                    request_id={self.exc.request_id})"



class MaxRetriesExhausted(Exception):
    """
    Max Retries Exhausted
    """

    def __init__(self, name: str, func_name: str, model: str, attempts: int, max_attempts: int):
        self.name = name
        self.func_name = func_name
        self.model = model
        self.attempts = attempts
        self.max_attempts = max_attempts
    
    def __str__(self):
        return f"[{self.name} - {self.model}] All retries exhausted for {self.func_name}. ({self.attempts}/{self.max_attempts})"
        
    def __repr__(self):
        """
        Returns the string representation of the error.
        """
        return f"[{self.name} - {self.model}] (func_name={self.func_name}, attempts={self.attempts}, max_attempts={self.max_attempts})"
        

class APIError(OpenAIError):
    pass


class TryAgain(OpenAIError):
    pass


class Timeout(OpenAIError):
    pass


class APIConnectionError(OpenAIError):
    pass


class InvalidRequestError(OpenAIError):
    pass


class InvalidAPIResourceError(OpenAIError):
    pass


class AuthenticationError(OpenAIError):
    pass


class PermissionError(OpenAIError):
    pass


class RateLimitError(OpenAIError):
    
    def post_init(self, **kwargs):
        """
        Gets the rate limit reset time
        """
        self.retry_after_seconds: Optional[float] = None
        with contextlib.suppress(Exception):
            self.retry_after_seconds = float(self.exc.error_message.split("Please retry after", 1)[1].split("second", 1)[0].strip())

class ServiceTimeoutError(OpenAIError):
    pass


class ServiceUnavailableError(OpenAIError):
    pass


class InvalidAPIType(OpenAIError):
    pass


class InvalidMaxTokens(InvalidRequestError):
    pass

    def post_init(self, **kwargs):
        """
        Gets the maximum context length and requested max tokens
        """
        self.maximum_context_length: Optional[int] = None
        self.requested_max_tokens: Optional[int] = None
        with contextlib.suppress(Exception):
            self.maximum_context_length = int(self.exc.error_message.split("maximum context length is", 1)[1].split(" ", 1)[0].strip())
            self.requested_max_tokens = int(self.exc.error_message.split("requested", 1)[1].split(" ", 1)[0].strip())


def fatal_exception(exc) -> bool:
    """
    Checks if the exception is fatal.
    """
    print(f"Checking if exception is fatal: {exc} ({type(exc)} = ({type(exc) == aiohttpx.ReadTimeout} is readtimeout)) ")

    if isinstance(exc, aiohttpx.ReadTimeout) or type(exc) == aiohttpx.ReadTimeout:
        return True
    
    if not isinstance(exc, OpenAIError):
        # retry on all other errors (eg. network)
        return False
    
    # retry on server errors and client errors
    # with 429 status code (rate limited),
    # with 400, 404, 415 status codes (invalid request),
    # 400 can include invalid parameters, such as invalid `max_tokens`
    # don't retry on other client errors
    if isinstance(exc, (InvalidMaxTokens, InvalidRequestError, MaxRetriesExhausted)):
        return True
    
    # Retry on invalid resource errors
    if isinstance(exc, InvalidAPIResourceError):
        return False
    
    return (400 <= exc.status < 500) and exc.status not in [429, 400, 404, 415, 524] # [429, 400, 404, 415]


def error_handler(
    response: aiohttpx.Response,
    data: Optional[Any] = None, # Line specific for streaming responses
    should_retry: Optional[bool] = False,
    **kwargs
):
    # sourcery skip: assign-if-exp, hoist-similar-statement-from-if, reintroduce-else, swap-nested-ifs

    if response.status_code == 503:
        raise ServiceUnavailableError(
            response = response,
            message = "The server is overloaded or not ready yet.",
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 429:
        raise RateLimitError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code in [400, 404, 415]:
        if 'maximum context length' in response.text:
            return InvalidMaxTokens(
                response = response,
                data = data,
                should_retry = False,
                **kwargs
            )
        if 'The API deployment for this resource does not exist' in response.text:
            return InvalidAPIResourceError(
                response = response,
                data = data,
                should_retry = should_retry,
                **kwargs
            )

        return InvalidRequestError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 401:
        return AuthenticationError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 403:
        return PermissionError(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    if response.status_code == 409:
        return TryAgain(
            response = response,
            data = data,
            should_retry = should_retry,
            **kwargs
        )
    
    # Service is likely down.
    if response.status_code == 524:
        raise ServiceTimeoutError(
            response = response,
            data = data,
            should_retry = False,
            **kwargs
        )
    
    raise APIError(
        response = response,
        data = data,
        should_retry = should_retry,
        **kwargs
    )



class MaxRetriesExceeded(Exception):
    def __init__(
        self,
        attempts: int,
        base_exception: OpenAIError,
        name: Optional[str] = None,
    ):
        self.name = name
        self.attempts = attempts
        self.ex = base_exception
    
    def __str__(self):
        return f"[{self.name}] Max {self.attempts} retries exceeded: {str(self.ex)}"
        
        
    @property
    def user_message(self):
        """
        Returns the error message.
        """
        return f"[{self.name}] Max {self.attempts} retries exceeded: {self.ex.user_message}"
    
    def __repr__(self):
        """
        Returns the string representation of the error.
        """
        return f"[{self.name}] {repr(self.ex)} (attempts={self.attempts})"
        


from __future__ import annotations

"""
Kinde OAuth2 Client
"""

from ..types.config import BaseOAuth2Config, eproperty
from ..types.client import BaseOAuth2Client
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated
from pydantic import Field
from lzo.utils import create_unique_id

_default_scopes = [
    'openid',
    'profile',
    'email',
    'offline',
]

class KindeOAuth2Config(BaseOAuth2Config):
    """
    Kinde Auth Settings
    """
    name: Optional[str] = 'kinde'
    scopes: Optional[List[str]] = Field(default = _default_scopes)

    class Config(BaseOAuth2Config.Config):
        env_prefix = "KINDE_"

    @eproperty
    def m2m_audience(self) -> str:
        """
        Returns the m2m audience
        """
        return f'{self.issuer_url}/api'

    def get_auth_params(
        self,
        redirect_url: str,
        scope: Optional[str] = None,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the auth params
        """
        scope = scope or self.get_scopes()
        state = state or create_unique_id(alph_only=True)
        return {
            'response_type': 'code',
            'client_id': self.client_id,
            'redirect_uri': redirect_url,
            'scope': scope,
            'state': state,
            'code_challenge': self.create_code_challenge(self.code_verifier),
            'code_challenge_method': self.code_challenge_method,
        }
    
    def get_fetch_token_params(
        self,
        code: str,
        redirect_url: str,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the fetch token params
        """
        return {
            'code': code,
            'code_verifier': self.code_verifier,
            'state': state,
            'redirect_uri': redirect_url,
            'grant_type': 'authorization_code',
        }
    
    def get_refresh_token_params(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the refresh token params
        """
        return {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
        }

    def get_headers(
        self,
        endpoint_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns the headers
        """
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json',
        }
        if endpoint_type == 'token':
            headers['Authorization'] = f'Basic {self.encode_client_secret_basic(self.client_id, self.client_secret)}'
        return headers


class KindeOAuth2Client(BaseOAuth2Client):
    """
    Kinde OAuth2 Client
    """
    name: Optional[str] = 'kinde'

    if TYPE_CHECKING:
        config: KindeOAuth2Config

    @eproperty
    def config_class(self) -> Type['KindeOAuth2Config']:
        """
        Returns the Config Class
        """
        return KindeOAuth2Config


    async def app_retrieve_existing_callbacks(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app callbacks
        """
        response = await self.m2m_client.async_get(
            f'/api/v1/applications/{self.config.client_id}/auth_redirect_urls',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        data: Dict[str, List[str]] = response.json()
        # logger.info(data)
        return data.get('redirect_urls', [])


    async def app_add_authorized_callbacks(
        self,
        callback_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized callbacks
        """
        response = await self.m2m_client.async_post(
            f'/api/v1/applications/{self.config.client_id}/auth_redirect_urls',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8',
            },
            json = {
                'urls': callback_urls,
            }
        )
        response.raise_for_status()
        data = response.json()
        data['urls'] = callback_urls
        self.logger.info(data, colored = True, prefix = f'{self.name} - Added {len(callback_urls)} Authorized Callbacks')
        
    
    async def app_retrieve_existing_logouts(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app logout urls
        """
        response = await self.m2m_client.async_get(
            f'/api/v1/applications/{self.config.client_id}/auth_logout_urls',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        data: Dict[str, List[str]] = response.json()
        return data.get('logout_urls', [])

    async def app_add_authorized_logouts(
        self,
        logout_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized logouts
        """
        response = await self.m2m_client.async_post(
            f'/api/v1/applications/{self.config.client_id}/auth_logout_urls',
            headers = {
                'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8',
            },
            json = {
                'urls': logout_urls,
            }
        )
        response.raise_for_status()
        data = response.json()
        data['urls'] = logout_urls
        self.logger.info(data, colored = True, prefix = f'{self.name} - Added {len(logout_urls)} Authorized Logouts')
    

    async def afetch_user_info(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ):
        """
        Fetches the user info
        """
        identity = token_payload.get('sub')
        if not identity: return
        response = await self.m2m_client.async_get(
            f'/api/v1/user?id={identity}',
            headers = {
                'Accept': 'application/json',
            }
        )
        response.raise_for_status()
        return response.json()

    async def afetch_access_token_info(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ):
        """
        Fetches the access token info
        """
        pass

    async def abuild_oauth_token(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Builds the oauth token
        """
        # The token should generally already be an access token
        if token_payload.get('at_hash'):
            return {
                'id_token': {
                    'value': token,
                    **self.parse_jwt_token(token, kind = 'id_token')
                }
            }
        return {
            'access_token': {
                'value': token,
                **self.parse_jwt_token(token, kind = 'access_token')
            }
        }


        